(******************************************************************************
 * © Mathias Bourgoin, Université Pierre et Marie Curie (2011)
 *
 * Mathias.Bourgoin@gmail.com
 *
 * This software is a computer program whose purpose is allow GPU programming 
 * with the OCaml language.
 *
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 ******************************************************************************)
open Vector
  
open Devices
  
open Mem
  
type kernel

external relax_vcustom : 'a vcustom -> 'b vcustom = "%identity"
  
type ('a, 'b) kernelArgs =
  | VCustom of 'a vcustom
  | VFloat32 of vfloat32
  | VFloat64 of vfloat64
  | VComplex32 of vcomplex32
  | VInt32 of vint32
  | VInt64 of vint64
  | Int32 of int
  | Int64 of int
  | Float32 of float
  | Float64 of float
  | Vector of ('a, 'b) vector
  | LocalFloat32 of localvfloat32

type block =
  { mutable blockX : int; mutable blockY : int; mutable blockZ : int
  }

type grid = { mutable gridX : int; mutable gridY : int; mutable gridZ : int }

let nbCaches = 64
  
module Cuda =
  struct
    type cuda_extra
    
    external cuda_create_extra : int -> cuda_extra = "spoc_cuda_create_extra"
      
    external cuda_compile :
      string ->
        string ->  generalInfo -> kernel =
      "spoc_cuda_compile"
      
    external cuda_load_param_vec :
      int ref -> cuda_extra -> device_vec -> unit =
      "spoc_cuda_load_param_vec_b" "spoc_cuda_load_param_vec_n"
      
    external cuda_custom_load_param_vec :
      int ref -> cuda_extra -> device_vec -> unit =
      "spoc_cuda_custom_load_param_vec_b" "spoc_cuda_custom_load_param_vec_n"
      
    external cuda_load_param_int : int ref -> cuda_extra -> int -> unit =
      "spoc_cuda_load_param_int_b" "spoc_cuda_load_param_int_n"
      
    external cuda_load_param_int64 : int ref -> cuda_extra -> int -> unit =
      "spoc_cuda_load_param_int64_b" "spoc_cuda_load_param_int64_n"
      
    external cuda_load_param_float : int ref -> cuda_extra -> float -> unit =
      "spoc_cuda_load_param_float_b" "spoc_cuda_load_param_float_n"
      
    external cuda_load_param_float64 :
      int ref -> cuda_extra -> float -> unit =
      "spoc_cuda_load_param_float64_b" "spoc_cuda_load_param_float64_n"
      
    external set_block_shape : kernel -> block -> generalInfo -> unit =
      "spoc_cuda_set_block_shape"
      
    external cuda_launch_grid :
      int ref -> kernel -> grid -> block -> cuda_extra -> generalInfo -> unit =
      "spoc_cuda_launch_grid_b" "spoc_cuda_launch_grid_n"
      
    external cuda_create_dummy_kernel : unit -> kernel =
      "spoc_cuda_create_dummy_kernel"
     
    let cudaKernelCache = ref [|  |]
      
    (*Array.make cuda_compatible_devices (Hashtbl.create nbCaches) *)
    let cuda_load cached id s1 s2 args =
      let path = Sys.argv.(0) in
      let kernelPath =
        (Filename.dirname path) ^ (Filename.dir_sep ^ (s1 ^ ".ptx"))
      in
        if cached
        then
          (if (Array.length !cudaKernelCache) = 0
           then
             cudaKernelCache :=
               Array.make !cuda_compatible_devices (Hashtbl.create nbCaches)
           else ();
           (let hash = Digest.file kernelPath
            in
              try let ker = Hashtbl.find !cudaKernelCache.(id) hash in ker
              with
              | Not_found ->
                  let src = ref "" in
                  let ic = open_in kernelPath
                  in
                    ((try
                        while true do src := !src ^ ((input_line ic) ^ "\n")
                          done
                      with | End_of_file -> ());
                     close_in ic;
                     let ker = cuda_compile !src s2
                     in (Hashtbl.add !cudaKernelCache.(id) hash ker; ker))))
        else
          (let src = ref "" in
           let ic = open_in kernelPath
           in
             ((try while true do src := !src ^ ((input_line ic) ^ "\n") done
               with | End_of_file -> ());
              close_in ic;
              let ker = cuda_compile !src s2 in ker))
      
    let cuda_load_arg offset extra dev cuFun idx vect =
      let load_non_vect =
        function
        | Int32 i -> cuda_load_param_int offset extra i
        | Int64 i -> cuda_load_param_int64 offset extra i
        | Float32 f -> cuda_load_param_float offset extra f
        | Float64 f -> cuda_load_param_float64 offset extra f
        | _ -> failwith "CU LOAD ARG Type Not Implemented\n"

      and check_vect v =
        (if !auto
         then
           (if v.device <> dev.general_info.id
            then Mem.to_device v dev
            else ();
            flush dev ~queue_id: 0 ())
         else ();
         match vect with
         | VCustom v ->
             cuda_custom_load_param_vec offset extra
               v.cuda_device_vec.(dev.general_info.id)
         | _ ->
             cuda_load_param_vec offset extra
               v.cuda_device_vec.(dev.general_info.id))
      in
        match vect with
        | VFloat32 v -> check_vect v
        | VFloat64 v -> check_vect v
        | VComplex32 v -> check_vect v
        | VInt32 v -> check_vect v
        | VInt64 v -> check_vect v
        | VCustom (v : ('a, 'a) vector) -> check_vect v
        | _ -> load_non_vect vect
      
  end
  
module OpenCL =
  struct
    external opencl_compile : string -> string -> generalInfo -> kernel =
      "spoc_opencl_compile"
      
    external opencl_compile_debug : string -> string -> generalInfo -> kernel =
      "spoc_debug_opencl_compile"
      
    external opencl_load_param_vec :
      int ref -> kernel -> int -> device_vec -> generalInfo -> unit =
      "spoc_opencl_load_param_vec"
      
    external opencl_load_param_local_vec :
      int ref -> kernel -> int -> device_vec -> generalInfo -> unit =
      "spoc_opencl_load_param_local_vec"
      
    external opencl_load_param_int :
      int ref -> kernel -> int -> generalInfo -> unit =
      "spoc_opencl_load_param_int"
      
    external opencl_load_param_int64 :
      int ref -> kernel -> int -> generalInfo -> unit =
      "spoc_opencl_load_param_int64"
      
    external opencl_load_param_float :
      int ref -> kernel -> float -> generalInfo -> unit =
      "spoc_opencl_load_param_float"
      
    external opencl_load_param_float64 :
      int ref -> kernel -> float -> generalInfo -> unit =
      "spoc_opencl_load_param_float64"
      
    external opencl_launch_grid :
      kernel -> grid -> block -> generalInfo -> int -> unit =
      "spoc_opencl_launch_grid"
      
    external opencl_create_dummy_kernel : unit -> kernel =
      "spoc_opencl_create_dummy_kernel"
      
    let openCLKernelCache = ref [|  |]
      
    (*Array.make opencl_compatible_devices (Hashtbl.create nbCaches)*)
    let opencl_load cached debug id s1 s2 =
      let path = Sys.argv.(0) in
      let kernelPath =
        (Filename.dirname path) ^ (Filename.dir_sep ^ (s1 ^ ".cl"))
      in
        if cached
        then
          (if (Array.length !openCLKernelCache) = 0
           then
             openCLKernelCache :=
               Array.make !opencl_compatible_devices
                 (Hashtbl.create nbCaches)
           else ();
           (let hash = Digest.file kernelPath
            in
              try let ker = Hashtbl.find !openCLKernelCache.(id) hash in ker
              with
              | Not_found ->
                  let src = ref "" in
                  let ic = open_in kernelPath
                  in
                    ((try
                        while true do src := !src ^ ((input_line ic) ^ "\n")
                          done
                      with | End_of_file -> ());
                     close_in ic;
                     let ker = opencl_compile !src s2
                     in (Hashtbl.add !openCLKernelCache.(id) hash ker; ker))))
        else
          (let src = ref "" in
           let ic = open_in kernelPath
           in
             ((try while true do src := !src ^ ((input_line ic) ^ "\n") done
               with | End_of_file -> ());
              close_in ic;
              let ker =
                if debug
                then opencl_compile_debug !src s2
                else opencl_compile !src s2
              in ker))
      
    let opencl_load_arg offset dev clFun idx vect =
      let load_non_vect =
        function
        | Int32 i -> opencl_load_param_int offset clFun i dev.general_info
        | Int64 i -> opencl_load_param_int64 offset clFun i dev.general_info
        | Float32 f ->
            opencl_load_param_float offset clFun f dev.general_info
        | Float64 f ->
            opencl_load_param_float64 offset clFun f dev.general_info
        | _ -> failwith "Cl LOAD ARG Type Not Implemented\n"

      and check_vect v =
        (if !auto
         then
           (if v.dev <> (Dev dev) then Mem.to_device v dev else ();
            flush dev ~queue_id: 0 ())
         else ();
         opencl_load_param_vec offset clFun idx
           v.opencl_device_vec.
             (dev.general_info.id - !cuda_compatible_devices)
           dev.general_info)
      in
        match vect with
        | VFloat32 v -> check_vect v
        | VFloat64 v -> check_vect v
        | VComplex32 v -> check_vect v
        | VInt32 v -> check_vect v
        | VInt64 v -> check_vect v
        | VCustom v -> check_vect v
        | LocalFloat32 v ->
            opencl_load_param_local_vec offset clFun idx
              v.opencl_device_vec.
                (dev.general_info.id - !cuda_compatible_devices)
              dev.general_info
        | _ -> load_non_vect vect
      
  end
  
open Cuda
  
open OpenCL
  
exception ERROR_BLOCK_SIZE
  
exception ERROR_GRID_SIZE
  
let exec  args (block, grid) cached debug queue_id
         dev bin =
  match dev.general_info.specific with
  | Cuda ->
      
       if !auto then () else ();
       let cuFun = bin in
       let offset = ref 0 in
       let extra = cuda_create_extra (Array.length args)
       in
         ((match dev.specific_info with
           | CudaInfo cI ->
               (if
                  (block.blockX > cI.maxThreadsDim.x) ||
                    ((block.blockY > cI.maxThreadsDim.y) ||
                       (block.blockZ > cI.maxThreadsDim.z))
                then raise ERROR_BLOCK_SIZE
                else ();
                if
                  (grid.gridX > cI.maxGridSize.x) ||
                    ((grid.gridY > cI.maxGridSize.y) ||
                       (grid.gridZ > cI.maxGridSize.z))
                then raise ERROR_GRID_SIZE
                else ())
           | _ -> failwith "Cuda Device without Cudainfo");
          Array.iteri (cuda_load_arg offset extra dev cuFun) args;
          (*set_block_shape cuFun block dev.general_info;*)
          cuda_launch_grid offset cuFun grid block extra dev.general_info)
  | OpenCL ->
       let clFun = bin in
       let offset = ref 0
       in
         (Array.iteri (opencl_load_arg offset dev clFun) args;
          opencl_launch_grid clFun grid block dev.general_info queue_id)
  | _ -> failwith "EXEC Not Implemented"
  
(*let compile (dev:device) ?debug:(d=false) ?cached:(c = false) ker =
	 match dev.general_info.specific with
  	| Cuda -> cuda_load c dev.general_info.id (snd (snd ker)) (fst (snd  ker)) (snd (fst ker))
           dev.general_info
		| OpenCL ->  opencl_load c d (dev.general_info.id - !cuda_compatible_devices)
           (snd (snd ker)) (fst (snd  ker)) dev.general_info
		| _ -> failwith "compile error"
*)
let compile_and_run (dev : device) ((block : block), (grid : grid))
                    ?cached:(c = false) ?debug:(d = false) ?queue_id:(q = 0)
                    ker =
  snd (fst ker) (block, grid) c d q dev
  
(*let run (dev:device) ((block:block),(grid:grid)) ?queue_id:(q = 0) ker args = 
	(fst (fst ker)) args (block,grid) q dev
 *)
let max a b = if a < b then b else a
  
let load_source file ext =
  let kernelPath =
		let path = Sys.argv.(0) in
    (Filename.dirname path) ^ (Filename.dir_sep ^ (file ^ ext)) in
  let src = ref "" in
  let ic = open_in kernelPath
  in
    ((try while true do src := !src ^ ((input_line ic) ^ "\n") done
      with | End_of_file -> ());
     close_in ic;
     !src)

exception No_source_for_device of Devices.device
exception Not_compiled_for_device of Devices.device

class  virtual ['a] spoc_kernel file (func:string) =
object (self)
	val file_file = file
	val kernel_name = func
	val mutable cuda_sources =
		try [ load_source file ".ptx" ] with | _ -> []
	
	val mutable opencl_sources =
		try [ load_source file ".cl" ] with | _ -> []
	
	val binaries = Hashtbl.create 8
	
	
	
	method get_cuda_sources = cuda_sources
	method set_cuda_sources s = 
		cuda_sources <- [s]

	method get_opencl_sources = opencl_sources
	method set_opencl_sources s = 
		opencl_sources <- [s]

	method reload_sources () =
		cuda_sources <- 
		(try [ load_source file ".ptx" ] with | _ -> []);
		opencl_sources <- 
		(try [ load_source file ".cl" ] with | _ -> [])
		
	method compile =
		fun dev ->
				try Hashtbl.find binaries dev
				with
				| Not_found ->
						let bin =
							(match dev.Devices.general_info.Devices.specific with
								| Devices.Cuda -> 
									begin
										match cuda_sources with
										| [] -> raise (No_source_for_device dev)
										| t::q -> cuda_compile t kernel_name dev.general_info
									end
								| Devices.OpenCL -> 
									begin
										match opencl_sources with
										| []  ->  raise (No_source_for_device dev)
										| t::q  -> opencl_compile t kernel_name dev.general_info
								end
								)
						in (Hashtbl.add binaries dev bin;
							bin)
	
	
	method virtual exec : 'a -> (block * grid) ->
           bool -> bool -> int -> Devices.device -> kernel  -> unit
	
	method run (args:'a) ((block:block), (grid:grid)) (cached:bool) (debug:bool) (queue_id:int) (dev:Devices.device) =
		let bin =
			try Hashtbl.find binaries dev with
				| Not_found -> 
					(
						try self#compile dev with
							| _  -> raise (Not_compiled_for_device dev)
					) 
		in
		self#exec args (block, grid) cached debug queue_id dev bin
		
	method compile_and_run (args:'a) ((block:block), (grid:grid)) (cached:bool) (debug:bool) (queue_id:int) (dev:Devices.device) =
		let bin = self#compile dev
		in 
		 self#exec args (block, grid) cached debug queue_id dev bin
	
end
  
	(*let run = compile_and_run*)
	let run dev (block,grid) k args = k#run args (block,grid) false false 0 dev

	let compile dev k = k#compile dev
