(******************************************************************************
 * © Mathias Bourgoin, Université Pierre et Marie Curie (2011)
 *
 * Mathias.Bourgoin@gmail.com
 *
 * This software is a computer program whose purpose is allow GPU programming 
 * with the OCaml language.
 *
 * This software is governed by the CeCILL-B license under French law and
 * abiding by the rules of distribution of free software.  You can  use, 
 * modify and/ or redistribute the software under the terms of the CeCILL-B
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info". 
 * 
 * As a counterpart to the access to the source code and  rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty  and the software's author,  the holder of the
 * economic rights,  and the successive licensors  have only  limited
 * liability. 
 * 
 * In this respect, the user's attention is drawn to the risks associated
 * with loading,  using,  modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean  that it is complicated to manipulate,  and  that  also
 * therefore means  that it is reserved for developers  and  experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or 
 * data to be ensured and,  more generally, to use and operate it in the 
 * same conditions as regards security. 
 * 
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL-B license and that you accept its terms.
 ******************************************************************************)

open Camlp4.PreCast
open Syntax
open Ast

exception Unauthorized_Recursion of string
exception Unauthorized_String_Access of string
exception Unauthorized_Array_Access of string
exception Unauthorized_Type of (string*string)
exception Unbound_value of (Loc.t*string)
exception Unbound_value2 of ((Loc.t*string)*Loc.t)

(******************** internal kernels **********************)

let expr_of_string = Syntax.Gram.parse_string Syntax.expr_eoi;;


let print i = 
		Printf.printf "%s\n" i; Pervasives.flush stdout

let print i = ()

type kexpr =
	| ExVAss of (loc*expr*expr)
	| ExLetM of (loc*binding*expr)


let rec string_of_ident i = 
	let aux = function
  | <:ident< $lid:s$ >> -> s
  | <:ident< $uid:s$ >> -> s
  | <:ident< $i1$.$i2$ >> -> "acc_" ^ (string_of_ident i1) ^ "_" ^ (string_of_ident i2)
  | <:ident< $i1$($i2$) >> -> "app_" ^ (string_of_ident i1) ^ "_" ^ (string_of_ident i2)
  | <:ident< $anti:_$ >> -> assert false
	in aux i
 

let new_vars = Hashtbl.create 10 


let used_vars = Hashtbl.create 10


let reset_vars () = 
	Hashtbl.clear new_vars;
	Hashtbl.clear used_vars;
	Hashtbl.add new_vars "0" Loc.ghost
	
	
let expr_of_kexpr kexpr =
	let rec aux = function
		| ExVAss (l,e0,e1)  -> ExAss (l,e0,e1)
	in aux kexpr
		

let keys () = 
	let tbl = Hashtbl.create 10 in
	Hashtbl.add tbl "acc_Spoc_acc_Kernel_dimX" "0";
	Hashtbl.add tbl "acc_Spoc_acc_Kernel_global_thread_idx" "get_global_id";
	Hashtbl.add tbl "float32" "float";
	Hashtbl.add tbl "float64" "double";
	tbl


let functions () = 
	let tbl = Hashtbl.create 10 in
	Hashtbl.add tbl "acc_Spoc_acc_Kernel_global_thread_idx" ("get_global_id",1);
	Hashtbl.add tbl "exit" ("return",0);
	Hashtbl.add tbl "float32" ("(float)",1);
	Hashtbl.add tbl "float64" ("(double)",1);
	Hashtbl.add tbl "mod" ("%",2);
	tbl

let param_types () = 
	let tbl = Hashtbl.create 10 in
	Hashtbl.add tbl "mutable" "__global";
	Hashtbl.add tbl "acc_acc_Spoc_Vector_vfloat32" "float*";
	Hashtbl.add tbl "acc_acc_Spoc_Vector_vfloat64" "double*";
	Hashtbl.add tbl "acc_acc_Spoc_Vector_vint32" "int*";
	Hashtbl.add tbl "acc_acc_Spoc_Vector_vint64" "long long*";	
	Hashtbl.add tbl "float32" "float";
	Hashtbl.add tbl "float64" "double";
	Hashtbl.add tbl "int" "int";
	tbl
	
let functions = functions ()

let keywords = keys ()

let param_types = param_types ()

let check_keyword t =
	try
		Hashtbl.find keywords (string_of_ident t)
	with Not_found  -> (*Printf.printf "%s not a keyword\n" (string_of_ident t);*) (string_of_ident t)


let is_function f =
	try
		(true,(Hashtbl.find functions (string_of_ident f)))
	with Not_found  -> (false,((string_of_ident f),0))
	
let check_param_types t loc=
	try
		Hashtbl.find param_types (string_of_ident t)
	with Not_found  -> raise (Unauthorized_Type ((Loc.to_string loc),string_of_ident t))
	


let rec parse_typ ctyp b =
	let rec aux b = function 
(*	  | TyNil of loc
    | TyAli of loc and ctyp and ctyp (* t as t *) (* list 'a as 'a *)
    | TyAny of loc (* _ *)*)
    | TyApp (l,t0,t1)  -> 
			(aux b t0)^" "^(aux false t1)
(*    | TyArr of loc and ctyp and ctyp (* t -> t *) (* int -> string *)
    | TyCls of loc and ident (* #i *) (* #point *)
    | TyLab of loc and string and ctyp (* ~s:t *)*)
    | TyId  (l,id) -> print "TyId"; 
			if b then 
				(
					if (check_param_types id l) <> "__global" then
						"const "^ (check_param_types id l)
					else
					(check_param_types id l)
			)
			else
				(check_param_types id l)
(*    | TyMan of loc and ctyp and ctyp (* t == t *) (* type t = [ A | B ] == Foo.t *)
      (* type t 'a 'b 'c = t constraint t = t constraint t = t *)
    | TyDcl of loc and string and list ctyp and ctyp and list (ctyp * ctyp)
      (* < (t)? (..)? > *) (* < move : int -> 'a .. > as 'a  *)
    | TyObj of loc and ctyp and row_var_flag
    | TyOlb of loc and string and ctyp (* ?s:t *)
    | TyPol of loc and ctyp and ctyp (* ! t . t *) (* ! 'a . list 'a -> 'a *)
    | TyQuo of loc and string (* 's *)
    | TyQuP of loc and string (* +'s *)
    | TyQuM of loc and string (* -'s *)
    | TyVrn of loc and string (* `s *)
    | TyRec of loc and ctyp (* { t } *) (* { foo : int ; bar : mutable string } *)
    | TyCol of loc and ctyp and ctyp (* t : t *)
    | TySem of loc and ctyp and ctyp (* t; t *)
    | TyCom of loc and ctyp and ctyp (* t, t *)
    | TySum of loc and ctyp (* [ t ] *) (* [ A of int and string | B ] *)
    | TyOf  of loc and ctyp and ctyp (* t of t *) (* A of int *)
    | TyAnd of loc and ctyp and ctyp (* t and t *)
    | TyOr  of loc and ctyp and ctyp (* t | t *)
    | TyPrv of loc and ctyp (* private t *)
    | TyMut of loc and ctyp (* mutable t *)
    | TyTup of loc and ctyp (* ( t ) *) (* (int * string) *)
    | TySta of loc and ctyp and ctyp (* t * t *)
    | TyVrnEq of loc and ctyp (* [ = t ] *)
    | TyVrnSup of loc and ctyp (* [ > t ] *)
    | TyVrnInf of loc and ctyp (* [ < t ] *)
    | TyVrnInfSup of loc and ctyp and ctyp (* [ < t > t ] *)
    | TyAmp of loc and ctyp and ctyp (* t & t *)
    | TyOfAmp of loc and ctyp and ctyp (* t of & t *)
    | TyPkg of loc and module_type (* (module S) *)
    | TyAnt of loc and string (* $s$ *) *)
		| _  -> "/*parse_typ: not implemented yet*/"
	in
		aux b ctyp

and params_to_object params _loc=
	let rec aux = function
		| ExApp (l,e0,e1) -> TySta (l, aux e0, aux e1)
		| ExTyc (l, e, ctyp) -> 
			begin
				match ctyp with
				| TyApp (l,t0,t1)  -> 
					begin
						begin
							match t0 with
							|  TyId (l,id) when (string_of_ident id = "mutable__")  -> 
								t1
							| t  -> t
				 		end;
						begin
							match t1 with
							|  TyId (l,id) when (string_of_ident id = "mutable__")  -> 
								t0
							| t  -> t
				 		end;
					end
				| t  ->  t
			end
		| _  -> failwith "params_to_object"
	in 
	TyTup (_loc, aux params)


and gen_fun_params prms _loc =
	let rec aux = function
		| ExApp (l,e0,e1) -> PaCom (l, aux e0, aux e1)
		| ExTyc (l, e, ctyp) -> 
			begin
				let id0 = 
					match e with
						| ExId (l,id)  -> PaId (l,id)
						| _  -> failwith "not an id"
				in
				match ctyp with
					| TyApp (l,t0,t1)  -> 
						begin
							begin
								match t0 with
								|  TyId (l,id) when (string_of_ident id = "mutable__")  -> 
									PaTyc (l,id0,t1)
								| t  -> PaTyc (l,id0,t)
				 		end;
						begin
							match t1 with
							|  TyId (l,id) when (string_of_ident id = "mutable__")  -> 
								PaTyc (l,id0,t0)
							| t  ->  PaTyc (l,id0,t)
				 		end;
					end
				| t  -> PaTyc (l,id0,t)
			end
		| _  -> failwith "params_to_object"
	in PaTup (_loc, aux prms)

	

	
	
	
and parse_params params = 
	let rec aux = function
		| ExNil l  -> print "ExNil"; ""
    | ExId  (l,id)  -> print "ExId"; (string_of_ident id)
    | ExAcc (l,e0,e1) -> print "ExAcc"; (aux e0)^(aux e1)
    | ExAnt (l,s)  -> print "ExAnt";  s
    | ExApp (l,e0,e1) -> print "ExApp"; (aux e0)^", "^(aux e1)
(*    | ExAre of loc and expr and expr (* e.(e) *)
    | ExArr of loc and expr (* [| e |] *)
    | ExSem of loc and expr and expr (* e; e *)
    | ExAsf of loc (* assert False *)
    | ExAsr of loc and expr (* assert e *)
    | ExAss of loc and expr and expr (* e := e *)
    | ExChr of loc and string (* 'c' *)
    | ExCoe of loc and expr and ctyp and ctyp (* (e : t) or (e : t :> t) *)
    | ExFlo of loc and string (* 3.14 *)
      (* for s = e to/downto e do { e } *)
    | ExFor of loc and string and expr and expr and direction_flag and expr
    | ExFun of loc and match_case (* fun [ mc ] *)
    | ExIfe of loc and expr and expr and expr (* if e then e else e *)
    | ExInt of loc and string (* 42 *)
    | ExInt32 of loc and string
    | ExInt64 of loc and string
    | ExNativeInt of loc and string
    | ExLab of loc and string and expr (* ~s or ~s:e *)
    | ExLaz of loc and expr (* lazy e *)
      (* let b in e or let rec b in e *)
    | ExLet of loc and rec_flag and binding and expr
      (* let module s = me in e *)
    | ExLmd of loc and string and module_expr and expr
      (* match e with [ mc ] *)
    | ExMat of loc and expr and match_case
      (* new i *)
    | ExNew of loc and ident
      (* object ((p))? (cst)? end *)
    | ExObj of loc and patt and class_str_item
      (* ?s or ?s:e *)
    | ExOlb of loc and string and expr
      (* {< rb >} *)
    | ExOvr of loc and rec_binding
      (* { rb } or { (e) with rb } *)
    | ExRec of loc and rec_binding and expr
      (* do { e } *)
    | ExSeq of loc and expr
      (* e#s *)
    | ExSnd of loc and expr and string
      (* e.[e] *)
    | ExSte of loc and expr and expr
      (* s *) (* "foo" *)
    | ExStr of loc and string
      (* try e with [ mc ] *)
    | ExTry of loc and expr and match_case
      (* (e) *)
    | ExTup of loc and expr
      (* e, e *)
    | ExCom of loc and expr and expr
      (* (e : t) *)*)
    | ExTyc (l, e, ctyp)  -> print "ExTyc"; 
			(
				let var = aux e
			in
				(try 
					if Loc.strictly_before l (Hashtbl.find new_vars var) then
						Hashtbl.replace new_vars var l 
				with Not_found  -> 
					(print ("add "^var^" to new_vars"); 
					Hashtbl.add new_vars var l)
				);
				parse_typ ctyp true^" "^(var)
				)
      (* `s *)
(*    | ExVrn of loc and string
      (* while e do { e } *)
    | ExWhi of loc and expr and expr
      (* let open i in e *)
    | ExOpI of loc and ident and expr
      (* fun (type t) -> e *)
      (* let f x (type t) y z = e *)
    | ExFUN of loc and string and expr
      (* (module ME : S) which is represented as (module (ME : S)) *)
    | ExPkg of loc and module_expr ] *)
		| _  ->  "/*parse_params: not implemented yet*/" 
		
		in 
			aux params


and parse_patt patt = 
	let rec aux = function
		| PaNil l -> print "PaNil"; ""
   	| PaId  (l,id) -> 
					(	print "PaId";  
					let var = (check_keyword id) in
					(try 
					if Loc.strictly_before l (Hashtbl.find new_vars var) then
						Hashtbl.replace new_vars var l 
					with Not_found  -> 	
						(
						print ("add "^var^" to new_vars");
						Hashtbl.add new_vars var l)
						);
						var
					)
(*    | PaAli of loc and patt and patt (* p as p *) (* (Node x y as n) *)
    | PaAnt of loc and string (* $s$ *)
    | PaAny of loc (* _ *)
    | PaApp of loc and patt and patt (* p p *) (* fun x y -> *)
    | PaArr of loc and patt (* [| p |] *)
    | PaCom of loc and patt and patt (* p, p *)
    | PaSem of loc and patt and patt (* p; p *)
    | PaChr of loc and string (* c *) (* 'x' *)
    | PaInt of loc and string
    | PaInt32 of loc and string
    | PaInt64 of loc and string
    | PaNativeInt of loc and string
    | PaFlo of loc and string
    | PaLab of loc and string and patt (* ~s or ~s:(p) *)
    (* ?s or ?s:(p) *)
    | PaOlb of loc and string and patt
    (* ?s:(p = e) or ?(p = e) *)
    | PaOlbi of loc and string and patt and expr
    | PaOrp of loc and patt and patt (* p | p *)
    | PaRng of loc and patt and patt (* p .. p *)
    | PaRec of loc and patt (* { p } *)
    | PaEq  of loc and ident and patt (* i = p *)
    | PaStr of loc and string (* s *)
    | PaTup of loc and patt (* ( p ) *)*)
    | PaTyc (l,p,t)  -> 
				(
					print "PaTyc"; 
					let var = (aux p) in
					(try 
					if Loc.strictly_before l (Hashtbl.find new_vars var) then
						Hashtbl.replace new_vars var l 
					with Not_found  -> 	
						(
						print ("add "^var^" to new_vars");
						Hashtbl.add new_vars var l)
						);
					(parse_typ t false^" "^var)
				)
(*    | PaTyp of loc and ident (* #i *)
    | PaVrn of loc and string (* `s *)
    | PaLaz of loc and patt (* lazy p *)
    | PaMod of loc and string (* (module M) *)*)
		| _  -> "/*parse_patt: not implemented yet*/"
	in
		aux patt

and parse_binding binding =
	let rec aux = function 
		| BiNil l  -> print "BiNil"; ""
      (* bi and bi *) (* let a = 42 and c = 43 *)
    | BiAnd (l,b0,b1)  ->  print "BiAnd";  (aux b0)^";\n"^(aux b1)
		  (* p = e *) (* let patt = expr *)
    | BiEq  (l,p,expr)  -> print "BiEq";
			begin
			(*	match expr with
				| ExApp (l,e0,e1) -> 
					begin
						match e0 with
						| ExId (l0,id) when (check_keyword id) = "ref"  ->
							(parse_patt p)^" = "^(parse_body e1) 
						| _ -> 
								"const "^(parse_patt p)^" = "^(parse_body expr)
   				end
				|_  -> *)
				let var = (parse_patt p) in
					"const "^var^" = "^(parse_body expr)
			end
		 | BiAnt (l,s)  -> print "BiAnt";  assert false
	
	in
		aux binding

and parse_mutable_binding binding =
	let rec aux = function 
		| BiNil l  -> print "BiNil"; ""
      (* bi and bi *) (* let a = 42 and c = 43 *)
    | BiAnd (l,b0,b1)  ->  print "BiAnd";  (aux b0)^";\n"^(aux b1)
		  (* p = e *) (* let patt = expr *)
    | BiEq  (l,p,expr)  -> print "BiEq";
			begin
			(*	match expr with
				| ExApp (l,e0,e1) -> 
					begin
						match e0 with
						| ExId (l0,id) when (check_keyword id) = "ref"  ->
							(parse_patt p)^" = "^(parse_body e1) 
						| _ -> 
								(parse_patt p)^" = "^(parse_body expr)
   				end
				|_  ->*) (parse_patt p)^" = "^(parse_body expr)
			end
		 | BiAnt (l,s)  -> print "BiAnt";  assert false
	
	in
		aux binding

and parse_bool bool =
	let rec aux = function
		| ExApp (l,e0,e1) -> print "ExApp"; 
					begin
						match e0 with
						| ExId (l0,id) when (check_keyword id) = "<"  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (parse_body e2_0)^" < "^(parse_body e2_1)
								| ExId (l,id) -> (parse_body e1)^" < "
								| _  -> "/*parse_body ExApp /:  not implemented yet*/"
							end; 
						| _  -> 
						"("^(aux e0)^(parse_body e1)^")"
					end
		|_ ->  "/*parse_bool: not implemented yet*/" 
	in aux bool

and parse_body body = 
	let rec aux = function
		| ExNil l  -> print "ExNil"; ""
    | ExId  (l,id)  -> print "ExId"; 
		  let var = (check_keyword id) in
				(try 
					if Loc.strictly_before l (Hashtbl.find used_vars var) then
						Hashtbl.replace used_vars var l 
					with Not_found  -> 	
						(
						print ("add "^var^" to used_vars");
						Hashtbl.add used_vars var l)
						);
					var
    | ExAcc (l,e0,e1) -> print "ExAcc"; (aux e0)^(aux e1)
    | ExAnt (l,s)  -> print "ExAnt";  s
    | ExApp (l,e0,e1) -> print "ExApp"; 
					begin
						match e0 with
						| ExId (l0,id) when ((check_keyword id) = "/" || (check_keyword id) = "/.")  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" / "
								| ExId (l,id) -> (aux e1)^" / "
								| _  -> (aux e1)^" / "
							end;
						| ExId (l0,id) when ((check_keyword id) = "*" || (check_keyword id) = "*.")
						  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" * "
								| ExId (l,id) -> (aux e1)^" * "
								| _  -> (aux e1)^" * " 
							end
						| ExId (l0,id) when ( (check_keyword id) = "+" || (check_keyword id) = "+.") -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" + "
								| _ -> (aux e1)^" + "
								| _  -> (aux e1)^" * "
							end
					| ExId (l0,id) when ( (check_keyword id) = "-" || (check_keyword id) = "-." ) -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" - "
								| ExId (l,id) -> (aux e1)^" - "
								| _  -> (aux e1)^" - "
								end
								
					| ExId (l0,id) when (check_keyword id) = "<"  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" < "
								| ExId (l,id) -> (aux e1)^" < "
								| _  -> (aux e1)^" < "
							end
					| ExId (l0,id) when (check_keyword id) = "<="  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" <= "
								| ExId (l,id) -> (aux e1)^" <= "
								| _  -> (aux e1)^" <= "
							end
					| ExId (l0,id) when (check_keyword id) = ">"  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" > "
								| ExId (l,id) -> (aux e1)^" > "
								| _  -> (aux e1)^" > "
							end
					| ExId (l0,id) when (check_keyword id) = ">="  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" >= "
								| ExId (l,id) -> (aux e1)^" >= "
								| _  -> (aux e1)^" >= "
							end			
					| ExId (l0,id) when (check_keyword id) = "="  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" == "
								| ExId (l,id) -> (aux e1)^" == "
								| _  -> (aux e1)^" == "
							end	
					| ExId (l0,id) when (check_keyword id) = "<>"  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" != "
								| ExId (l,id) -> (aux e1)^" != "
								| _  -> (aux e1)^" != "
							end			

					| ExId (l0,id) when (check_keyword id) = "&&"  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" && "
								| ExId (l,id) -> (aux e1)^" && "
								| _  -> "/*parse_body ExApp &&:  not implemented yet*/"
							end	
					| ExId (l0,id) when (check_keyword id) = "||"  -> 
							begin
								match e1 with 
								| ExApp  (l1,e2_0,e2_1) -> (aux e1)^" || "
								| ExId (l,id) -> (aux e1)^" || "^ (aux e1)
								| _  -> "/*parse_body ExApp ||:  not implemented yet*/"
							end			
									
													
					| ExId (l0,id)  ->
						let r =  is_function id in
						if (fst r) then
							begin
								if (snd (snd r)) = 0 then
								(fst (snd r))^";\n"
								else
								if (snd (snd r)) = 1 then
								(fst (snd r))^" ("^(aux e1)^")"
								else
								if (snd (snd r)) = 2 then
									begin
									if (fst (snd r)) = "%" then
										" ("^(aux e1)^")"^(fst (snd r))
									else
										assert false
									end
								else
								 assert false
							end
						else "("^(aux e0)^(aux e1)^")"
					| _  -> 
						"("^(aux e0)^(aux e1)^")"
					end	
    | ExAre (l,e0,e1)  -> print "ExAre"; (aux e0)^"["^(aux e1)^"]"
(*    | ExArr of loc and expr (* [| e |] *)*)
    | ExSem (l,e0,e1)  -> print "ExSem"; (aux e0)^"\n"^(aux e1)
(*    | ExAsf of loc (* assert False *)
    | ExAsr of loc and expr (* assert e *)*)
    | ExAss (l,e0,e1)  ->  print "ExAss"; 
		let var = (aux e0) in
			var^" = "^(aux e1)^";"
(*    | ExChr of loc and string (* 'c' *)
    | ExCoe of loc and expr and ctyp and ctyp (* (e : t) or (e : t :> t) *)*)
      (* for s = e to/downto e do { e } *)
		| ExFlo (l,s)  -> print "ExFlo"; s
(*    | ExFor of loc and string and expr and expr and direction_flag and expr
    | ExFun of loc and match_case (* fun [ mc ] *)*)
    | ExIfe (l,e0,e1,e2)  ->  
			begin
				match e2 with
				| ExNil l  -> "if ("^(aux e0)^") {\n"^
																(aux e1)^"}\n"
				| ExId (l,id) when (string_of_ident id = "()")  -> "if ("^(aux e0)^") {\n"^
																(aux e1)^"}\n"
				| _  -> "if ("^(aux e0)^") {\n"^
																(aux e1)^"}\nelse {"^(aux e2)^"}\n"
			end;
    | ExInt (l,s)  -> print "ExInt"; s
(*    | ExInt32 of loc and string
    | ExInt64 of loc and string
    | ExNativeInt of loc and string
    | ExLab of loc and string and expr (* ~s or ~s:e *)
    | ExLaz of loc and expr (* lazy e *)
      (* let b in e or let rec b in e *) *)
    | ExLet (l,r,b,expr)  ->  print "ExLet";
			begin
				match r with 
					| ReNil  -> 	
							(parse_binding b)^";\n"^(aux expr)
					| ReRecursive  -> (parse_mutable_binding b)^";\n"^(aux expr)
					| _   -> assert false
			end;
      (* let module s = me in e *)
(*    | ExLmd of loc and string and module_expr and expr
      (* match e with [ mc ] *)
			.
    | ExMat of loc and expr and match_case
      (* new i *)
    | ExNew of loc and ident
      (* object ((p))? (cst)? end *)
    | ExObj of loc and patt and class_str_item
      (* ?s or ?s:e *)
    | ExOlb of loc and string and expr
      (* {< rb >} *)
    | ExOvr of loc and rec_binding
      (* { rb } or { (e) with rb } *)
    | ExRec of loc and rec_binding and expr
      (* do { e } *)*)
    | ExSeq (l,expr)  -> print "ExSeq"; (aux expr)
      (* e#s *)
(*    | ExSnd of loc and expr and string
      (* e.[e] *)
    | ExSte of loc and expr and expr
      (* s *) (* "foo" *)
    | ExStr of loc and string
      (* try e with [ mc ] *)
    | ExTry of loc and expr and match_case
      (* (e) *)
    | ExTup of loc and expr
      (* e, e *)
    | ExCom of loc and expr and expr
      (* (e : t) *)*)
    | ExTyc (l, e, ctyp)  -> print "ExTyc"; (parse_typ ctyp false)^" "^(aux e)
      (* `s *)
(*    | ExVrn of loc and string
      (* while e do { e } *) *)
    | ExWhi (l,e0,e1)  -> print "ExWhi"; 
			"while ("^(aux e0)^") {\n"^(aux e1)^"\n}"
      (* let open i in e *)
(*    | ExOpI of loc and ident and expr
      (* fun (type t) -> e *)
      (* let f x (type t) y z = e *)
    | ExFUN of loc and string and expr
      (* (module ME : S) which is represented as (module (ME : S)) *)
    | ExPkg of loc and module_expr ] *)
		| _  ->  "/*parse_body: not implemented yet*/" 
		
		in 
			aux body

let evaluate_diff s1 s2 = 
	let r = ref 0. in
			let div = ref 1 in
	if (
			(
				(String.length s1) <=  
				(
					(String.length s2) + 1 + 
							((String.length s2) *10)/100
				)) 
				
				&& 
				(
					(String.length s1) >=  
						(
							(String.length s2) - 1 -  
								((String.length s2) *10)/100) 
				)
		) then
		begin
		
				for i = 0 to (String.length s1) - 1  do
					if i < String.length s2 then
						if s1.[i] = s2.[i] then
							r := ! r +. 1.
						else
							for j = 0 to !div do
								if i + j < String.length s2 then
								(
									if s1.[i] = s2.[i+j] then
									(	
										r:= !r +. 1./. (float !div);
										incr div
									)
								);
								if i - j > 0 then
								(
									if s1.[i] = s2.[i-j] then
									(
										r:= !r +. 1./. (float !div);
										incr div
									)
								);
							done;
			done;
		end;
		!r
(*		if (int_of_float !r) <= (String.length s2 + 1 + (String.length s2 *10)/100) && (int_of_float !r) >=  (String.length s2 - 1 -  (String.length s2 *10)/100) 	then
			true
		else 
			(
				print_float !r;
				print_newline ();
			false
*)

let check_vars () =
	try 
		Hashtbl.iter (fun v l -> 
		try 
			if Loc.strictly_before 
			l  
			(Hashtbl.find new_vars v) then
				raise (Unbound_value2 ((l, v), (Hashtbl.find new_vars v)))
			with 
				| Not_found  -> raise (Unbound_value (l, v)) ) used_vars
		with	
			| Unbound_value2  ((l,v),l2) -> failwith ("Value '"^v^"' not bound here ("^Loc.to_string l^"), will be bound here ("^(Loc.to_string l2)^")" )
			| Unbound_value (unbound_l, unbound_v)  -> 
				(
					let res = ref [] 
					and final_res = ref [] in
					Hashtbl.iter ( fun v l  -> 
					if Loc.strictly_before l unbound_l then
						let diff = evaluate_diff v unbound_v in
						if  diff >= 0. then
							res := [(l,v),diff]@ !res;
					)new_vars;
					let rec aux res = 
						match res with
							| t::q  ->
								begin
									match !final_res with
										| []  -> final_res := [t]
										| final_t::q2  -> 
											if snd t > snd final_t then
											(	
												final_res := [t];
											);
											
									end;
									aux q
							| []  ->
								begin
								match !final_res with
									| []  ->  failwith ("Value "^unbound_v^" not bound ("^Loc.to_string unbound_l^")")	
									| t::q  -> 
										failwith ("Value '"^unbound_v^
														"' not bound here ("^
														Loc.to_string unbound_l^
														"), did you mean '"
														^(snd (fst t))^
														"' ("^
														(Loc.to_string (fst (fst t)))^
														")")
								end
								in aux !res
							)
									
						
let inkern = ref false

let gen_opencl_code _loc kernel_name prms body =
	reset_vars ();
	let name = (string_of_ident kernel_name)
	and header = "__kernel void "
	and params = (parse_params prms)
	and body = (parse_body body)
	and kernel_name = <:ident< $lid:("opencl_"^(string_of_ident kernel_name))$ >>
	in
	let str = header^
						name^
						(*params*)"( "^params^" )"^
						"{\n"^body^"}"
	and f = <:patt< $id:kernel_name$ >>	
	in
		let expr = ExStr (_loc, str)
				in
					check_vars ();
					inkern := false;
					<:str_item< let $f$ = $expr$ >>

let gen_cuda_code  _loc kernel_name prms body =
	reset_vars ();
	let name = (string_of_ident kernel_name)
	and header = "__kernel void "
	and params = (parse_params prms)
	and body = (parse_body body)
	and kernel_name = <:ident< $lid:("cuda_"^(string_of_ident kernel_name))$ >>	
	in

	let str = header^
						name^
						(*params*)"( "^params^" )"^
						"{\n"^body^"}"
	and f = <:patt< $id:kernel_name$ >>	
	in
		let expr = ExStr (_loc, str)
				in
					check_vars ();
					inkern := false;
					<:str_item< let $f$ = $expr$ >>








 let bigarray_set _loc var newval =
    match var with
    <:expr< Bigarray.Array1.get $arr$ $c1$ >> ->
        Some <:expr< Bigarray.Array1.set $arr$ $c1$ $newval$ >>
    | <:expr< Bigarray.Array2.get $arr$ $c1$ $c2$ >> ->
        Some <:expr< Bigarray.Array2.set $arr$ $c1$ $c2$ $newval$ >>
    | <:expr< Bigarray.Array3.get $arr$ $c1$ $c2$ $c3$ >> ->
        Some <:expr< Bigarray.Array3.set $arr$ $c1$ $c2$ $c3$ $newval$ >>
    | <:expr< Bigarray.Genarray.get $arr$ [| $coords$ |] >> ->
        Some <:expr< Bigarray.Genarray.set $arr$ [| $coords$ |] $newval$ >>
		| <:expr< Spoc.Mem.get $arr$ $c1$ >>  -> 
		Some <:expr< Spoc.Mem.set $arr$ $c1$ $newval$ >>
    | _ -> None 

let conc_seq e1 e2 =
    match (e1, e2) with
    |(ExSeq(l,e1),ExSeq(l2,e2)) ->
       ExSeq(l, ExSem(l,e1,e2))
    | (ExSeq(l,e1),_)->
       ExSeq(l, ExSem(l,e1,e2)) 
    | (_,ExSeq(l,e2)) ->
        ExSeq(l, ExSem(l,e1,e2))
    | _ ->
      (
				  let _loc =
          Loc.merge (Ast.loc_of_expr e1)
                    (Ast.loc_of_expr e2) in
											ExSeq(_loc, ExSem(_loc,e1,e2))
        
				)
				
(******************** external kernels **********************)

let expr_of_string = Syntax.Gram.parse_string Syntax.expr_eoi;;


let id = ref 0



let arg_string = ref ""
let arg_string2 = ref ""

let idx = ref 1

let create_new_var loc =
	let v = "spoc_var"^(string_of_int !id) 
	in 
	let a = expr_of_string loc (v) in
	incr id;
	a,v


let rec string_of_ident = function
  | <:ident< $lid:s$ >> -> s
  | <:ident< $uid:s$ >> -> s
  | <:ident< $i1$.$i2$ >> -> "acc_" ^ (string_of_ident i1) ^ "_" ^ (string_of_ident i2)
  | <:ident< $i1$($i2$) >> -> "app_" ^ (string_of_ident i1) ^ "_" ^ (string_of_ident i2)
  | <:ident< $anti:_$ >> -> assert false

let fst3 = fun (x,y,z) -> x
let snd3 = fun (x,y,z) -> y
let thrd3 = fun (x,y,z) -> z

let rec parse_ctyp = function
	| TyNil _ -> failwith "TyNil";
	| TyAli _ -> failwith "TyAli";
	| TyAny _ -> failwith "TyAny";
	| TyApp (l, c1, c2) as t -> 
		let e,v =create_new_var l in
		let l1 = parse_ctyp c1 
		and l2 =parse_ctyp c2 in
		[ExTyc (l, e, t)],[v],[t] 
		(*failwith "TyApp";*)
	| TyArr (l,a,b) -> 
		(*Printf.printf "TyArr\n";*) 
		let l1 = parse_ctyp a 
		and l2 =parse_ctyp b in
			((fst3 l1)@(fst3 l2),(snd3 l1)@(snd3 l2),(thrd3 l1)@(thrd3 l2))
	| TyCls _ -> failwith "TyCls";
	| TyLab _ -> failwith "TyLab";
	| (TyId (l,id)) as t -> let e,v =create_new_var l in
				begin
					match t with
					| TyId(_, id) when (String.compare (string_of_ident id) "int32") = 0
						-> ([ExTyc (l,e,TyId(l, ident_of_expr (expr_of_string l "int")))],[v],[t])
					| TyId(_, id) when (String.compare (string_of_ident id) "int64") = 0
							-> ([ExTyc (l,e,TyId(l, ident_of_expr (expr_of_string l "int")))],[v],[t])
					| TyId(_, id) when (String.compare (string_of_ident id) "float32") = 0
						-> ([ExTyc (l,e,TyId(l, ident_of_expr (expr_of_string l "float")))],[v],[t])
					| TyId(_, id) when (String.compare (string_of_ident id) "float64") = 0
							-> ([ExTyc (l,e,TyId(l, ident_of_expr (expr_of_string l "float")))],[v],[t])		
					| _ ->  ([ExTyc (l,e,t)],[v],[t])
				end;
	| _ -> failwith "parse_ctyp"
	
	
let translate_type = function
	| TyId(l, id) when (String.compare (string_of_ident id) "int32") = 0
						-> TyId(l, ident_of_expr (expr_of_string l "int"))
	| TyId(l, id) when (String.compare (string_of_ident id) "int64") = 0
						-> TyId(l, ident_of_expr (expr_of_string l "int"))
	| TyId(l, id) when (String.compare (string_of_ident id) "float32") = 0
						-> TyId(l, ident_of_expr (expr_of_string l "float"))
	| TyId(l, id) when (String.compare (string_of_ident id) "float64") = 0
						-> TyId(l, ident_of_expr (expr_of_string l "float"))
	| t -> t

	
		
			
let rec parseTyp t i= 
						match t with
							| TyId (_, id) when (String.compare (string_of_ident id) "int") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Int32 "^i^"; "^(!arg_string); 
								arg_string2 :=i^" "^(!arg_string2)
							| TyId (_, id) when (String.compare (string_of_ident id) "int32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Int32 "^i^"; "^(!arg_string); 
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "int64") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Int64 "^i^"; "^(!arg_string);
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "float") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Float32 "^i^"; "^(!arg_string); 
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "float32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Float32 "^i^"; "^(!arg_string); 
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "float64") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.Float64 "^i^"; "^(!arg_string); 
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vfloat32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VFloat32 "^i^"; "^(!arg_string);
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_localvfloat32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.LocalFloat32 "^i^"; "^(!arg_string);
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vfloat64") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VFloat64 "^i^"; "^(!arg_string);  
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vcomplex32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VComplex32 "^i^"; "^(!arg_string);
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vint32") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VInt32 "^i^"; "^(!arg_string); 
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vint64") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VInt64 "^i^"; "^(!arg_string); 
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vbool") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VCustom  (Spoc.Kernel.relax_vcustom "^i^"); "^(!arg_string);
								arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vcustom") = 0 -> 
								incr idx;
								arg_string := "Spoc.Kernel.VCustom  (Spoc.Kernel.relax_vcustom "^i^"); "^(!arg_string); 
																arg_string2 :=i^" "^(!arg_string2)

							| TyId (_, id) -> 
								(*Printf.printf "--- %s\n" (string_of_ident id); 
								flush stdout;  *)
								arg_string := i^"; "^(!arg_string);
								arg_string2 := i^" "^(!arg_string2)

								(*failwith "apply_of_constraints TyId "*)
							| TyApp (l, t1, t2)  -> 
								begin
									match t2 with 
									| TyId (_, id) ->
										(*incr idx;*)
										(parseTyp t1 i );(*arg_string:= (string_of_ident id)^" "^ !arg_string;*)
									| _  -> failwith "erf"
								end
							| _ -> 
								failwith "apply_of_constraints Not TyId "				
					
						
							
								
									
										
											
												
														
let  gen_ktyp loc type_list = 
	let l = List.tl (List.rev type_list) in
		let l = List.rev l in
				TyTup (loc,	 Ast.tySta_of_list l)
(*				List.hd l*)
	
			

let ident_of_string loc s= IdLid (loc, s)

let gen_args loc type_list =
	id := 0;
	arg_string :=  "";
	let var_list = ref [] in
	let _loc = loc in
		let l = List.tl (List.rev type_list) in
				let l = List.rev l in
						let translate t = 
							let nv = create_new_var loc in
								var_list :=!var_list @ [nv];
									PaTyc (loc, 
													(PaId (loc,
													(ident_of_string loc (snd	nv))
													)), 
												t
									) 
		in
			let tuple =  Ast.paCom_of_list (List.map (translate) l) in
		arg_string := "";
		
		
		let array =	
			let arr_content =
				let f t = 
					arg_string := "";
 					parseTyp t (snd (List.nth !var_list !idx)) ;
  				expr_of_string loc !arg_string
				in 
				idx := 0 ;					
				Ast.exSem_of_list (List.map f l)
			in
			ExArr(loc, arr_content)
			in
			idx := 1 ;
		PaTup(loc, tuple),array
		
				


and gen_exec_params prms _loc =
	let rec internal_parse_typ e = function
		| TyId (l, id) when (String.compare (string_of_ident id) "int") = 0 -> 
			expr_of_string l ("Spoc.Kernel.Int32 "^(string_of_ident e)); 
		| TyId (l, id) when (String.compare (string_of_ident id) "int32") = 0 -> 
			expr_of_string l ("Spoc.Kernel.Int32 "^(string_of_ident e)); 
		| TyId (l, id) when (String.compare (string_of_ident id) "int64") = 0 -> 
			expr_of_string l ("Spoc.Kernel.Int64 "^(string_of_ident e));
		| TyId (l, id) when (String.compare (string_of_ident id) "float") = 0 -> 
			expr_of_string l ("Spoc.Kernel.Float32 "^(string_of_ident e));
		| TyId (l, id) when (String.compare (string_of_ident id) "float32") = 0 -> 
			expr_of_string l ("Spoc.Kernel.Float32 "^(string_of_ident e));
		| TyId (l, id) when (String.compare (string_of_ident id) "float64") = 0 -> 
			expr_of_string l ("Spoc.Kernel.Float64 "^(string_of_ident e));	
		| TyId (l, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vfloat32") = 0 -> 
			expr_of_string l ("Spoc.Kernel.VFloat32 "^(string_of_ident e))
		| TyId (l, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_localvfloat32") = 0 -> 
			expr_of_string l ("Spoc.Kernel.LocalFloat32 "^(string_of_ident e))
		| TyId (l, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vfloat64") = 0 -> 
			expr_of_string l ("Spoc.Kernel.VFloat64 "^(string_of_ident e))
		| TyId (l, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vcomplex32") = 0 -> 
			expr_of_string l ("Spoc.Kernel.VComplex32 "^(string_of_ident e))
		| TyId (l, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vint32") = 0 -> 
			expr_of_string l ("Spoc.Kernel.VInt32 "^(string_of_ident e))
		| TyId (l, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vint64") = 0 -> 
			expr_of_string l ("Spoc.Kernel.VInt64 "^(string_of_ident e))
		| TyId (l, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vbool") = 0 -> 
			expr_of_string l ("Spoc.Kernel.VBool "^(string_of_ident e))
		| TyId (l, id) when (String.compare (string_of_ident id) "acc_acc_Spoc_Vector_vcustom") = 0 -> 
			expr_of_string l ("Spoc.Kernel.VCustom  (Spoc.Kernel.relax_vcustom "^(string_of_ident e))
		| TyApp (l, t1, t2)  -> 
								begin
									match t2 with 
									| TyId (_, id) ->
										(*incr idx;*)
										(internal_parse_typ e t1  );(*arg_string:= (string_of_ident id)^" "^ !arg_string;*)
									| _  -> failwith "erf"
								end
		| TyId (l,id) ->  print (string_of_ident id);
				expr_of_string l (string_of_ident e)
		
	in
			let rec aux = function
		| ExApp (l,e0,e1) -> ExSem (l, aux e0, aux e1)
		| ExTyc (l, e, ctyp) -> 
			begin
				let e = match e with
					| ExId (l,id) -> id 
					| _  -> failwith "invalid parameter"
				in 		
				match ctyp with
					| TyApp (l,t0,t1)  -> 
						begin
							begin
								match t0 with
								|  TyId (l,id) when (string_of_ident id = "mutable__")  -> 
									(internal_parse_typ e t1)
								| t  -> (internal_parse_typ e t)
				 		end;
						begin
							match t1 with
							|  TyId (l,id) when (string_of_ident id = "mutable__")  -> 
								(internal_parse_typ e t0)
							| t  ->  (internal_parse_typ e t)
				 		end;
					end
				| t  -> (internal_parse_typ e t)
			end
		| _  -> failwith "params_to_object"
	in ExArr (_loc, aux prms)



let gen_internal_object _loc kernel_name prms = 
	let k_fun = PaId (_loc, IdLid (_loc, "kernel_"^(string_of_ident kernel_name))) 
	and cuda_sources = expr_of_string _loc  ("cuda_"^(string_of_ident kernel_name))
	and opencl_sources = expr_of_string _loc  ("opencl_"^(string_of_ident kernel_name))
(*			in
	let k_fun2 = <:patt< $k_fun$ >>*)
	in
		
		let exec_fun =
						<:binding< 
						$k_fun$ = fun $(gen_fun_params prms _loc)$  ->
						Spoc.Kernel.exec  $(gen_exec_params prms _loc)$
						>>
						 in
	
	let exec = ExId(_loc, IdLid (_loc, "kernel_"^(string_of_ident kernel_name))) in
	let p1 = PaId (_loc, kernel_name) in 
	let class_name = "class_kernel_"^(string_of_ident kernel_name) in
		let obj_name = ExId (_loc, kernel_name) 
		and meth_cu = expr_of_string _loc "set_cuda_sources"
		and	meth_cl = expr_of_string _loc "set_opencl_sources"
	  and kernel_exp_name = ExStr (_loc, string_of_ident kernel_name) 
in
	let method_call_cuda = 
	(string_of_ident kernel_name)^"#"^ "set_cuda_sources " ^"cuda_"^(string_of_ident kernel_name)

	
	and method_call_opencl = 
	(string_of_ident kernel_name)^"#"^ "set_opencl_sources " ^"opencl_"^(string_of_ident kernel_name)

	in
		let method_calls =
			StSem(_loc,
			StExp (_loc, expr_of_string _loc method_call_cuda),
			StExp (_loc, expr_of_string _loc method_call_opencl)
			)
	in
		let obj = <:str_item< 
		class $lid:class_name$  = object (self)
		inherit [$params_to_object prms _loc$] 
		Spoc.Kernel.spoc_kernel "" $kernel_exp_name$
		method exec  = $exec$
		end
		let $p1$ = new $lid:class_name$;;
		$method_calls$ 
		>>
		
		
	in	StSem(_loc,
							StVal	(_loc, ReNil,exec_fun),				
							obj)
							
(* array access *)
DELETE_RULE Gram expr: SELF; "."; "("; SELF; ")" END
(* string access *)
DELETE_RULE Gram expr: SELF; "."; "["; SELF; "]" END

(* recursion *)
DELETE_RULE Gram opt_rec: "rec" END


(*DELETE_RULE Gram sem_expr: expr LEVEL "top"; ";";SELF END
DELETE_RULE Gram sem_expr: expr LEVEL "top" ; ";" END
DELETE_RULE Gram sem_expr: expr LEVEL "top" END
DELETE_RULE Gram sequence: sem_expr END
DELETE_RULE Gram do_sequence: sequence; "done" END*)

DELETE_RULE Gram expr: "let"; opt_rec; binding; "in"; expr LEVEL ";" END

(* vector *)
(*DELETE_RULE Gram expr: SELF; "<-"; expr LEVEL "top" END*)


			

EXTEND Gram
  GLOBAL: str_item  sequence do_sequence expr ctyp opt_rec;
 	str_item: LEVEL "top"
	[
		["kernelet"; start_kern; kernel_name = ident; "="; "kern";   prms = expr; "->"; body = expr; end_kern-> 
				let _loc = Loc.ghost in
			StSem(_loc, StSem(_loc,
				gen_opencl_code _loc kernel_name prms body,
				gen_cuda_code _loc kernel_name prms body),
				gen_internal_object _loc kernel_name prms)	]
	];
	start_kern:
	[[ -> inkern := true;
	]];
	end_kern:
	[[ -> inkern := false;
	]];	
	(*sem_expr:
      [ [ e1 = expr LEVEL "top"; ";"; e2 = SELF -> <:expr< $e1$; $e2$ >>
        | e = expr LEVEL "top"; ";" -> e
        | e = expr LEVEL "top" -> e ] ]
    ;
    sequence:
      [ [ e = sem_expr -> e ] ]
    ;
    do_sequence:
      [ [ seq = sequence; "done" -> seq
      ] ]
    ;
		*)
		expr:  LEVEL "."
		[
		LEFTA
		[	e0 = SELF; "."; "[<"; e1 = SELF; ">]" ->
			if !inkern then
				 ExAre (_loc,e0,e1) 
			else
					let e = expr_of_string _loc "Spoc.Mem.get" in
						<:expr< $e$ $e0$ $e1$ >>
		|e0 = SELF; "."; "("; e1 = SELF; ")" -> 
			if !inkern then
			raise (Unauthorized_Array_Access (Loc.to_string _loc))
			else
				<:expr< $e0$ .( $e1$ ) >>
		|e0 = SELF; "."; "["; e1 = SELF; "]" -> if !inkern then
			raise (Unauthorized_String_Access (Loc.to_string _loc))
		else 
			<:expr< $e0$ .[ $e1$ ] >>
		]
	];
	seq_expr:
      [ [ e1 = expr LEVEL "top"; ";"; e2 = SELF ->
            conc_seq e1 e2
        | e1 = expr LEVEL "top"; ";" -> e1
        | e1 = expr LEVEL "top" -> e1 ] ];
	expr: BEFORE "top"
      [ ";" [ e = seq_expr -> e ] ];
	expr: LEVEL "top" 
	[
		["let"; "mutable"; b = binding; "in"; e = expr LEVEL ";"  -> 
					ExLet(_loc, ReRecursive, b,e)
(*		|	"let"; b = binding; "in"; e = SELF  -> 
					ExLet(_loc, ReNil, b,e)	*)
		|	"let"; r = opt_rec; bi = binding; "in"; x = expr LEVEL ";" ->
         if !inkern then
				match r with
					| ReNil  -> 	ExLet(_loc, ReNil, bi,x)
					| _  -> raise (Unauthorized_Recursion (Loc.to_string _loc))
					else
						ExLet (_loc,r,bi,x) 
		]
	];
	expr: LEVEL "top"
      [ 
				[ e1 = SELF; ":="; e2 = expr LEVEL "top" ->
            <:expr< $e1$ := $e2$ >>
				|
				 e1 = SELF; "<-"; e2 = expr LEVEL "top" ->
         match bigarray_set _loc e1 e2 with
          Some e -> e
         | None ->  ExAss (_loc, e1, e2)
      ] ];
	ctyp:
	[
		["mutable"; t = SELF  -> TyApp (_loc, TyId(_loc, IdLid(_loc,"mutable")),t)]
	];
	opt_rec:
	[
		["rec" -> if !inkern then
			raise (Unauthorized_Recursion (Loc.to_string _loc))
			else
				 <:rec_flag< rec >>]
	];
	
(******************** external kernels **********************)
	
	str_item: 
   [ [ "kernel"; lid = ident; ":";  typ = ctyp; "=";  file_name = STRING; func_name = STRING  -> 
					let constraint_var, arg_list, id_list  = parse_ctyp typ 
					and id =  IdLid  (_loc, "kernel_"^(string_of_ident lid) )in

					let p1 = PaId (_loc, lid) in 

					let idArgs = IdLid  (_loc, "args") in
					let f =  PaId (_loc,  idArgs) in
						
					let k_typ = gen_ktyp _loc id_list in
					let args = gen_args _loc id_list in
					let k_name = "kernel_"^(string_of_ident lid) in
					let k_fun = <:patt<
					 $id:id$
					>> 
					in
					let exec_fun =
						<:binding< 
						$k_fun$ = 
							
							fun  $fst args$ ->
							 Spoc.Kernel.exec $snd args$ 
						>> in
						
					let class_name = "class_kernel_"^(string_of_ident lid)
					and	filename =ExStr (_loc, file_name)
					and funcname =ExStr (_loc, func_name) in
					let exec = ExId(_loc, id) in
					
							let class_ident = <:ident< 	spoc_kernel>>
							in
							
						let gen_object = 
							<:str_item< 
								
								class $lid:class_name$  = object (self)
									inherit [$k_typ$ (*-> Spoc.Kernel.block * Spoc.Kernel.grid ->
         							bool -> bool -> int -> Spoc.Devices.device  ->  unit*)] 
												Spoc.Kernel.spoc_kernel $filename$ $funcname$
							method exec  = 
											$exec$

						end
								let $p1$ = new $lid:class_name$;;
								 >> in


						arg_string := "";
				
							StSem(_loc,
							StVal	(_loc, ReNil,exec_fun),				
							gen_object)
							]];
	
	END


